import scala.annotation.tailrec
import parser._

package object main {
  trait Exp
  // Aritmética
  case class Num(v: Double) extends Exp
  case class Soma(e1: Exp, e2: Exp) extends Exp
  case class Mult(e1: Exp, e2: Exp) extends Exp
  // Condicionais
  case class Menor(e1: Exp, e2: Exp) extends Exp
  case class If(econd: Exp, ethen: Exp, eelse: Exp) extends Exp
  case class True() extends Exp
  case class False() extends Exp
  // Usado apenas na avaliação verificada (stepc)
  case class Erro() extends Exp
  // Funções de primeira ordem
  case class Fun1(nome: String, params: List[String], corpo: Exp)
  case class Prog(funs: Set[Fun1], corpo: Exp)
  case class Var(nome: String) extends Exp
  case class Ap1(fun: String, args: List[Exp]) extends Exp
    
  trait Valor
  case class NumV(v: Double) extends Valor
  case class TrueV() extends Valor
  case class FalseV() extends Valor
  
  def val2exp(v: Valor): Exp = v match {
    case NumV(v) => Num(v)
    case TrueV() => True()
    case FalseV() => False()
  }
  
  def ProgFun: Parser[Prog] = for {
    funs <- many(DefFun)
    corpo <- ExpFun
  } yield Prog(funs.toSet, corpo)
    
  def DefFun: Parser[Fun1] = for {
    _ <- kw("fun")
    (nome, _) <- id
    _ <- op("(")
    params <- (for {
      (p, _) <- id
      ps <- many(for {
        _ <- op(",")
        (p, _) <- id
      } yield p)
    } yield p :: ps) +: empty(List())
    _ <- op(")")
    corpo <- ExpFun
    _ <- kw("end")
  } yield Fun1(nome, params, corpo) 
  
  def ExpFun: Parser[Exp] =
    chainl(AexpFun,
           for { _ <- op("<") } yield (e1, e2) => Menor(e1, e2),
           AexpFun)
  
  def AexpFun: Parser[Exp] =
    chainl(TermoFun, 
        (for { _ <- op("+") } yield (e1, e2) => Soma(e1, e2)) +:
        (for { _ <- op("-") } yield (e1: Exp, e2: Exp) => 
          Soma(e1, Mult(Num(-1), e2))),
        TermoFun)
    
  def TermoFun: Parser[Exp] =
    chainl(FatorFun, for {
      _ <- op("*")
    } yield (e1, e2) => Mult(e1, e2), FatorFun)
    
  def FatorFun: Parser[Exp] =
    (for { (v, pos) <- num } yield Num(v)) +:
    (for {
      _ <- op("(")
      e <- ExpFun
      _ <- op(")")
    } yield e) +:
    (for {
      _ <- op("-")
      (v, pos) <- num
    } yield Mult(Num(-1), Num(v))) +:
    (for {
      _ <- kw("if")
      econd <- ExpFun
      _ <- kw("then")
      ethen <- ExpFun
      _ <- kw("else")
      eelse <- ExpFun
      _ <- kw("end")
    } yield If(econd, ethen, eelse)) +:
    (for { _ <- kw("true") } yield True()) +:
    (for { _ <- kw("false") } yield False()) +:
    (for {
      (nome, _) <- id
      _ <- op("(")
      args <- (for {
        a <- ExpFun
        as <- many(for {
          _ <- op(",")
          a <- ExpFun
        } yield a)
      } yield a :: as) +: empty(List())
      _ <- op(")")
    } yield Ap1(nome, args)) +:
    (for { (nome, _) <- id } yield Var(nome))
    
  def fun: Parser[Prog] = for {
    p <- ProgFun
    _ <- space
    _ <- not(pred(c => true), ())
  } yield p
  
  def eval(p: Prog): Valor = p match {
    case Prog(funs, corpo) => eval(funs)(corpo)
  }
  
  def eval(funs: Set[Fun1])= {
    def eval(e: Exp): Valor = e match {
      case Num(v) => NumV(v)
      case Soma(e1, e2) => {
        val (NumV(v1), NumV(v2)) = (eval(e1), eval(e2))
        NumV(v1 + v2)
      }
      case Mult(e1, e2) => {
        val (NumV(v1), NumV(v2)) = (eval(e1), eval(e2))
        NumV(v1 * v2)
      }
      case Menor(e1, e2) => {
        val (NumV(v1), NumV(v2)) = (eval(e1), eval(e2))
        if(v1 < v2) TrueV() else FalseV()
      }
      case If(econd, ethen, eelse) => eval(econd) match {
        case TrueV() => eval(ethen)
        case FalseV() => eval(eelse)
      }
      case True() => TrueV()
      case False() => FalseV()
      case Ap1(nome, args) => {
        val vargs: List[Exp] = args.map(arg => val2exp(eval(arg)))
        val Fun1(_, params, corpo) = funs.find(f => f.nome == nome).get
        eval(subst(params.zip(vargs).toMap)(corpo))
      }
      // case Var(nome) => ??? expressão indefinida (erro!)
    }
    eval _  // retorna a função eval interna (com funs fixado)
  }
  
  def subst(n2v: Map[String,Exp]) = {
    def subst(e: Exp): Exp = e match {
      case Soma(e1, e2) => Soma(subst(e1), subst(e2))
      case Mult(e1, e2) => Mult(subst(e1), subst(e2))
      case Menor(e1, e2) => Menor(subst(e1), subst(e2))
      case If(ec, et, ee) => If(subst(ec), subst(et), subst(ee))
      case Ap1(nome, args) => Ap1(nome, args.map(arg => subst(arg)))
      case Var(nome) => n2v.get(nome).get
      case _ => e
    }
    subst _
  }
  
  def run(s: String): Valor = eval(fun.parse(s))
  
  def step(funs: Set[Fun1]) = {
    def step(e: Exp): Exp = e match {
      case Num(v)                    => Num(v)
      case Soma(Num(v1), Num(v2))    => Num(v1 + v2)
      case Soma(Num(v), e)           => Soma(Num(v), step(e))
      case Soma(e1, e2)              => Soma(step(e1), e2)
      case Mult(Num(v1), Num(v2))    => Num(v1 * v2)
      case Mult(Num(v), e)           => Mult(Num(v), step(e))
      case Mult(e1, e2)              => Mult(step(e1), e2)
      case True()                    => True()
      case False()                   => False()
      case Menor(Num(v1), Num(v2))   => if (v1 < v2) True() else False()
      case Menor(Num(v), e)          => Menor(Num(v), step(e))
      case Menor(e1, e2)             => Menor(step(e1), e2)
      case If(True(), ethen, eelse)  => ethen
      case If(False(), ethen, eelse) => eelse
      case If(econd, ethen, eelse)   => If(step(econd), ethen, eelse)
      case Ap1(nome, args)           => {
        if(args.exists(arg => arg.isInstanceOf[Soma] ||
                              arg.isInstanceOf[Mult] ||
                              arg.isInstanceOf[Menor] ||
                              arg.isInstanceOf[If] ||
                              arg.isInstanceOf[Ap1]))
          Ap1(nome, stepList(args))
        else {
          val Fun1(_, params, corpo) = funs.find(f => f.nome == nome).get
          subst(params.zip(args).toMap)(corpo)
        }
      }
      // case Var(nome)              => ??? indefinido (erro)
    }
    
    def stepList(args: List[Exp]): List[Exp] = args match {
      case Nil => Nil
      case Num(v) :: t => Num(v) :: stepList(t)
      case True() :: t => True() :: stepList(t)
      case False() :: t => False() :: stepList(t)
      case h :: t => step(h) :: t
    }
    
    step _
  }
    
  def evals(funs: Set[Fun1]) = {
    def evals(e: Exp): Valor = e match {
      case Num(v) => NumV(v)
      case True() => TrueV()
      case False() => FalseV()
      case e1 => {
        val e2 = step(funs)(e1)
        if(e1 == e2) error("travou " + e1) else evals(e2)
      }
    }
    evals _
  } 
  
  def evals(p: Prog): Valor = evals(p.funs)(p.corpo)
  
  def runs(s: String): Valor = evals(fun.parse(s))
}
